from pathlib import Path
from unittest.mock import Mock

from ftl_extract.code_extractor import sort_fluent_keys_by_path
from ftl_extract.matcher import FluentKey


def fluent_key_mock(path: str) -> FluentKey:
    mock = Mock(spec=FluentKey)
    mock.path = Path(path)
    return mock


def test_sort_fluent_keys_by_single_path() -> None:
    fluent_keys = {
        "key-1": fluent_key_mock("path/to/file1.ftl"),
    }
    expected = {
        Path("path/to/file1.ftl"): [fluent_keys["key-1"]],
    }
    assert sort_fluent_keys_by_path(fluent_keys) == expected


def test_sort_fluent_keys_by_multiple_paths() -> None:
    fluent_keys = {
        "key-1": fluent_key_mock("path/to/file1.ftl"),
        "key-2": fluent_key_mock("path/to/file2.ftl"),
    }
    expected = {
        Path("path/to/file1.ftl"): [fluent_keys["key-1"]],
        Path("path/to/file2.ftl"): [fluent_keys["key-2"]],
    }
    assert sort_fluent_keys_by_path(fluent_keys) == expected


def test_sort_fluent_keys_by_same_path() -> None:
    fluent_keys = {
        "key-1": fluent_key_mock("path/to/file.ftl"),
        "key-2": fluent_key_mock("path/to/file.ftl"),
    }
    expected = {
        Path("path/to/file.ftl"): [fluent_keys["key-1"], fluent_keys["key-2"]],
    }
    assert sort_fluent_keys_by_path(fluent_keys) == expected


def test_sort_fluent_keys_empty_dict() -> None:
    fluent_keys = {}
    expected = {}
    assert sort_fluent_keys_by_path(fluent_keys) == expected


def test_sort_fluent_keys_with_nonexistent_path() -> None:
    fluent_keys = {
        "key-1": fluent_key_mock("nonexistent/path.ftl"),
    }
    expected = {
        Path("nonexistent/path.ftl"): [fluent_keys["key-1"]],
    }
    assert sort_fluent_keys_by_path(fluent_keys) == expected



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from fluent.syntax import ast, parse

from ftl_extract.matcher import FluentKey

if TYPE_CHECKING:
    from fluent.syntax.ast import Resource


def import_from_ftl(
    path: Path,
    locale: str,
) -> tuple[dict[str, FluentKey], Resource, list[FluentKey]]:
    """Import `FluentKey`s from FTL."""
    ftl_keys = {}
    leave_as_is = []

    resource = parse(path.read_text(encoding="utf-8"), with_spans=True)

    for position, entry in enumerate(resource.body, start=0):
        if isinstance(entry, ast.Message):
            ftl_keys[entry.id.name] = FluentKey(
                code_path=Path(),
                key=entry.id.name,
                translation=entry,
                path=path,
                locale=locale,
                position=position,
            )
        else:
            leave_as_is.append(
                FluentKey(
                    code_path=Path(),
                    key="",
                    translation=entry,
                    path=path,
                    locale=locale,
                    position=position,
                ),
            )

    return ftl_keys, resource, leave_as_is


def import_ftl_from_dir(path: Path, locale: str) -> tuple[dict[str, FluentKey], list[FluentKey]]:
    """Import `FluentKey`s from directory of FTL files."""
    ftl_files = (path / locale).rglob("*.ftl") if path.is_dir() else [path]
    ftl_keys = {}
    leave_as_is = []

    for ftl_file in ftl_files:
        keys, _, as_is_keys = import_from_ftl(ftl_file, locale)
        ftl_keys.update(keys)
        leave_as_is.extend(as_is_keys)

    return ftl_keys, leave_as_is



from __future__ import annotations

import ast
from dataclasses import dataclass, field
from math import inf
from pathlib import Path
from typing import TYPE_CHECKING, cast

from fluent.syntax import ast as fluent_ast

from ftl_extract.const import (
    GET_LITERAL,
    I18N_LITERAL,
    IGNORE_ATTRIBUTES,
    IGNORE_KWARGS,
    PATH_LITERAL,
)
from ftl_extract.exceptions import (
    FTLExtractorDifferentPathsError,
    FTLExtractorDifferentTranslationError,
)

if TYPE_CHECKING:
    from collections.abc import Iterable


@dataclass
class FluentKey:
    """
    Dataclass for storing information about key and its translation.

    :param code_path: Path to .py file where key was found.
    :type code_path: Path
    :param key: Key that will be used to get translation.
    :type key: str | None
    :param translation: Translation of key.
    :type translation: str | None
    :param path: Path to .ftl file where key will be stored.
    :type path: Path
    :param locale: Locale of translation. When extracting from .py file, it will not be needed.
    :type locale: str | None
    """

    code_path: Path
    key: str
    translation: fluent_ast.EntryType
    path: Path
    locale: str | None = field(default=None)
    position: int | float = field(default=inf)


class I18nMatcher(ast.NodeVisitor):
    def __init__(
        self,
        code_path: Path,
        default_ftl_file: Path,
        func_names: str | Iterable[str] = I18N_LITERAL,
        ignore_attributes: str | Iterable[str] = IGNORE_ATTRIBUTES,
        ignore_kwargs: str | Iterable[str] = IGNORE_KWARGS,
    ) -> None:
        """

        :param code_path: Path to .py file where visitor will be used.
        :type code_path: Path
        :param default_ftl_file: Default name of FTL file.
        :type default_ftl_file: Path
        :param func_names: Name of function that is used to get translation. Default is "i18n".
        :type func_names: str | Iterable[str]
        :param ignore_attributes: Ignore attributes, like `i18n.set_locale`.
        :type ignore_attributes: str | Iterable[str]
        :param ignore_kwargs: Ignore kwargs, like `when` from
        `aiogram_dialog.I18nFormat(..., when=...)`.
        :type ignore_kwargs: str | Iterable[str]
        """
        self.code_path = code_path
        self.func_names = (
            frozenset({func_names}) if isinstance(func_names, str) else frozenset(func_names)
        )
        self.ignore_attributes = (
            frozenset({ignore_attributes})
            if isinstance(ignore_attributes, str)
            else frozenset(ignore_attributes)
        )
        self.ignore_kwargs = (
            frozenset({ignore_kwargs})
            if isinstance(ignore_kwargs, str)
            else frozenset(ignore_kwargs)
        )
        self.default_ftl_file = default_ftl_file
        self.fluent_keys: dict[str, FluentKey] = {}

    def visit_Call(self, node: ast.Call) -> None:  # noqa: N802
        # Check if the call matches the pattern
        if isinstance(node.func, ast.Attribute):
            attr: ast.Attribute | ast.expr = node.func
            attrs = []
            while isinstance(attr, ast.Attribute):
                attrs.append(attr.attr)
                attr = attr.value

            if isinstance(attr, ast.Name) and attr.id in self.func_names:
                if len(attrs) == 1 and attrs[0] == GET_LITERAL:
                    # Check if the call has args
                    if not node.args:
                        return  # Skip if no args

                    # Add the first arg as the translation key
                    attrs.clear()
                    if isinstance(arg := node.args[0], ast.Constant):
                        key = cast(ast.Constant, arg).value

                    else:
                        self.generic_visit(node)
                        return

                    fluent_key = create_fluent_key(
                        code_path=self.code_path,
                        key=key,
                        keywords=node.keywords,
                        ignore_kwargs=self.ignore_kwargs,
                        default_ftl_file=self.default_ftl_file,
                    )

                    process_fluent_key(self.fluent_keys, fluent_key)

                else:
                    if attrs[-1] in self.ignore_attributes:
                        self.generic_visit(node)
                        return

                    fluent_key = create_fluent_key(
                        code_path=self.code_path,
                        key="-".join(reversed(attrs)),
                        keywords=node.keywords,
                        ignore_kwargs=self.ignore_kwargs,
                        default_ftl_file=self.default_ftl_file,
                    )
                    process_fluent_key(self.fluent_keys, fluent_key)
            else:
                self.generic_visit(node)

        elif isinstance(node.func, ast.Name) and node.func.id in self.func_names:
            if not node.args:
                return

            fluent_key = create_fluent_key(
                code_path=self.code_path,
                key=cast(ast.Constant, node.args[0]).value,
                keywords=node.keywords,
                ignore_kwargs=self.ignore_kwargs,
                default_ftl_file=self.default_ftl_file,
            )
            process_fluent_key(self.fluent_keys, fluent_key)

        self.generic_visit(node)


def create_fluent_key(
    code_path: Path,
    key: str,
    keywords: list[ast.keyword],
    ignore_kwargs: frozenset[str],
    default_ftl_file: Path,
) -> FluentKey:
    fluent_key = FluentKey(
        code_path=code_path,
        key=key,
        translation=fluent_ast.Message(
            id=fluent_ast.Identifier(name=key),
            value=fluent_ast.Pattern(elements=[fluent_ast.TextElement(value=key)]),
        ),
        path=default_ftl_file,
    )

    for kw in keywords:
        if kw.arg == PATH_LITERAL:
            if kw.value is not None and isinstance(kw.value, ast.Constant):
                fluent_key.path = Path(cast(ast.Constant, kw.value).value)
        elif isinstance(kw.arg, str):
            if kw.arg in ignore_kwargs:
                continue

            cast(
                fluent_ast.Pattern,
                cast(fluent_ast.Message, fluent_key.translation).value,
            ).elements.append(
                fluent_ast.Placeable(
                    expression=fluent_ast.VariableReference(id=fluent_ast.Identifier(name=kw.arg)),
                ),
            )

    return fluent_key


def process_fluent_key(fluent_keys: dict[str, FluentKey], new_fluent_key: FluentKey) -> None:
    if new_fluent_key.key in fluent_keys:
        if fluent_keys[new_fluent_key.key].path != new_fluent_key.path:
            raise FTLExtractorDifferentPathsError(
                new_fluent_key.key,
                new_fluent_key.path,
                fluent_keys[new_fluent_key.key].path,
            )
        if not fluent_keys[new_fluent_key.key].translation.equals(new_fluent_key.translation):
            raise FTLExtractorDifferentTranslationError(
                new_fluent_key.key,
                cast(fluent_ast.Message, new_fluent_key.translation),
                cast(fluent_ast.Message, fluent_keys[new_fluent_key.key].translation),
            )

    else:
        fluent_keys[new_fluent_key.key] = new_fluent_key



from __future__ import annotations

from typing import TYPE_CHECKING

from fluent.syntax import ast

if TYPE_CHECKING:
    from ftl_extract.matcher import FluentKey


def _extract_kwargs_from_variable_reference(
    variable_reference: ast.VariableReference,
    kwargs: set[str],
) -> None:
    kwargs.add(variable_reference.id.name)


def _extract_kwargs_from_selector_expression(
    selector_expression: ast.SelectExpression,
    kwargs: set[str],
) -> None:
    if isinstance(selector_expression.selector, ast.VariableReference):
        _extract_kwargs_from_variable_reference(selector_expression.selector, kwargs)

    for variant in selector_expression.variants:
        for placeable in variant.value.elements:
            if isinstance(placeable, ast.Placeable):
                _extract_kwargs_from_placeable(placeable, kwargs)


def _extract_kwargs_from_placeable(placeable: ast.Placeable, kwargs: set[str]) -> None:
    expression = placeable.expression

    if isinstance(expression, ast.VariableReference):
        _extract_kwargs_from_variable_reference(expression, kwargs)

    elif isinstance(expression, ast.SelectExpression):
        _extract_kwargs_from_selector_expression(expression, kwargs)


def extract_kwargs(key: FluentKey) -> set[str]:
    kwargs: set[str] = set()

    if not isinstance(key.translation, ast.Message):
        return kwargs

    if not key.translation.value:
        return kwargs

    for placeable in key.translation.value.elements:
        if isinstance(placeable, ast.Placeable):
            _extract_kwargs_from_placeable(placeable, kwargs)

    return kwargs



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

    from fluent.syntax import ast


class FTLExtractorError(Exception):
    pass


class FTLExtractorDifferentPathsError(FTLExtractorError):
    def __init__(self, key: str, current_path: Path, new_path: Path) -> None:
        self.current_path = current_path
        self.new_path = new_path
        super().__init__(
            f"Key {key!r} already exists with different path: "
            f"{self.current_path} != {self.new_path}",
        )


class FTLExtractorDifferentTranslationError(FTLExtractorError):
    def __init__(
        self,
        key: str,
        current_translation: ast.Message,
        new_translation: ast.Message,
    ) -> None:
        self.current_translation = current_translation
        self.new_translation = new_translation
        super().__init__(
            f"Translation {key!r} already exists with different elements: "
            f"{self.current_translation} != {self.new_translation}",
        )



from __future__ import annotations

import ast
from pathlib import Path
from typing import TYPE_CHECKING, cast

from fluent.syntax import ast as fluent_ast

from ftl_extract.exceptions import (
    FTLExtractorDifferentPathsError,
    FTLExtractorDifferentTranslationError,
)
from ftl_extract.matcher import I18nMatcher

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from ftl_extract.matcher import FluentKey


def find_py_files(path: Path) -> Iterator[Path]:
    """
    First step: find all .py files in given path.

    :param path: Path to directory with .py files.
    :type path: Path
    :return: Iterator with Path to .py files.
    :rtype: Iterator[Path]
    """
    yield from path.rglob("[!{.}]*.py") if path.is_dir() else [path]


def parse_file(
    path: Path,
    i18n_keys: str | Iterable[str],
    ignore_attributes: str | Iterable[str],
    ignore_kwargs: str | Iterable[str],
    default_ftl_file: Path,
) -> dict[str, FluentKey]:
    """
    Second step: parse given .py file and find all i18n calls.

    :param path: Path to .py file.
    :type path: Path
    :param i18n_keys: Names of function that is used to get translation.
    :type i18n_keys: str | Iterable[str]
    :param ignore_attributes: Ignore attributes, like `i18n.set_locale`.
    :type ignore_attributes: str | Iterable[str]
    :param ignore_kwargs: Ignore kwargs, like `when` from
    `aiogram_dialog.I18nFormat(..., when=...)`.
    :type ignore_kwargs: str | Iterable[str]
    :param default_ftl_file: Default name of FTL file.
    :type default_ftl_file: Path
    :return: Dict with `key` and `FluentKey`.
    :rtype: dict[str, FluentKey]
    """
    node = ast.parse(path.read_bytes())
    matcher = I18nMatcher(
        code_path=path,
        default_ftl_file=default_ftl_file,
        func_names=i18n_keys,
        ignore_attributes=ignore_attributes,
        ignore_kwargs=ignore_kwargs,
    )
    matcher.visit(node)
    return matcher.fluent_keys


def post_process_fluent_keys(fluent_keys: dict[str, FluentKey], default_ftl_file: Path) -> None:
    """
    Third step: post-process parsed `FluentKey`.

    :param fluent_keys: Dict with `key` and `FluentKey` that will be post-processed.
    :type fluent_keys: dict[str, FluentKey]
    :param default_ftl_file: Default name of FTL file.
    :type default_ftl_file: Path
    """
    for fluent_key in fluent_keys.values():
        if not isinstance(fluent_key.path, Path):
            fluent_key.path = Path(fluent_key.path)

        if not fluent_key.path.suffix:  # if path looks like directory (no suffix)
            fluent_key.path /= default_ftl_file


def find_conflicts(
    current_fluent_keys: dict[str, FluentKey],
    new_fluent_keys: dict[str, FluentKey],
) -> None:
    """
    Fourth step: find conflicts between current and new `FluentKey`s.

    If conflict is found, raise `ValueError`.

    Conflict is when `key` is the same, but `path` or `kwargs` are different.
    """
    # Find common keys
    conflict_keys = set(current_fluent_keys.keys()) & set(new_fluent_keys.keys())

    if not conflict_keys:
        return

    for key in conflict_keys:
        if current_fluent_keys[key].path != new_fluent_keys[key].path:
            raise FTLExtractorDifferentPathsError(
                key,
                current_fluent_keys[key].path,
                new_fluent_keys[key].path,
            )

        if not current_fluent_keys[key].translation.equals(new_fluent_keys[key].translation):
            raise FTLExtractorDifferentTranslationError(
                key,
                cast(fluent_ast.Message, current_fluent_keys[key].translation),
                cast(fluent_ast.Message, new_fluent_keys[key].translation),
            )


def extract_fluent_keys(
    path: Path,
    i18n_keys: str | Iterable[str],
    ignore_attributes: str | Iterable[str],
    ignore_kwargs: str | Iterable[str],
    default_ftl_file: Path,
) -> dict[str, FluentKey]:
    """
    Extract all `FluentKey`s from given path.

    :param path: Path to [.py file] / [directory with .py files].
    :type path: Path
    :param i18n_keys: Names of function that is used to get translation.
    :type i18n_keys: str | Iterable[str]
    :param ignore_attributes: Ignore attributes, like `i18n.set_locale`.
    :type ignore_attributes: str | Iterable[str]
    :param ignore_kwargs: Ignore kwargs, like `when` from
    `aiogram_dialog.I18nFormat(..., when=...)`.
    :type ignore_kwargs: str | Iterable[str]
    :param default_ftl_file: Default name of FTL file.
    :type default_ftl_file: Path
    :return: Dict with `key` and `FluentKey`.
    :rtype: dict[str, FluentKey]

    """
    fluent_keys: dict[str, FluentKey] = {}

    for file in find_py_files(path):
        keys = parse_file(
            path=file,
            i18n_keys=i18n_keys,
            ignore_attributes=ignore_attributes,
            ignore_kwargs=ignore_kwargs,
            default_ftl_file=default_ftl_file,
        )
        post_process_fluent_keys(keys, default_ftl_file)
        find_conflicts(fluent_keys, keys)
        fluent_keys.update(keys)

    return fluent_keys


def sort_fluent_keys_by_path(fluent_keys: dict[str, FluentKey]) -> dict[Path, list[FluentKey]]:
    """
    Sort `FluentKey`s by their paths.

    :param fluent_keys: Dict with `key` and `FluentKey`.
    :type fluent_keys: dict[str, FluentKey]
    :return: Dict with `Path` and list of `FluentKey`.
    :rtype: dict[Path, list[FluentKey]]
    """
    sorted_fluent_keys: dict[Path, list[FluentKey]] = {}
    for fluent_key in fluent_keys.values():
        sorted_fluent_keys.setdefault(fluent_key.path, []).append(fluent_key)

    return sorted_fluent_keys



from pathlib import Path
from typing import Final

from ftl_extract.code_extractor import extract_fluent_keys
from ftl_extract.const import DEFAULT_FTL_FILE, IGNORE_ATTRIBUTES, IGNORE_KWARGS

CONTENT: Final[str] = """
def test(i18n):
    i18n.get("key-1")
    i18n.get("key-2", _path="content/file.ftl")
    i18n.get("key-3", arg_1="arg-1", arg_2="arg-2", _path="content/file.ftl")
    i18n.get("key-4", arg_1=arg1, arg_2=arg2)
    i18n.get("key-5", arg_1=obj.arg1, arg_2=obj.arg2)
    i18n.get("key-6", arg_1=obj.arg1(), arg_2=obj.arg2())

    i18n.sugar_key_one()
    i18n.sugar_key_two(_path="content/file.ftl")
    i18n.sugar_key_three(arg_1="arg-1", arg_2="arg-2", _path="content/file.ftl")
    i18n.sugar_key_four(arg_1=arg1, arg_2=arg2)
    i18n.sugar_key_five(arg_1=obj.arg1, arg_2=obj.arg2)
    i18n.sugar_key_six(arg_1=obj.arg1(), arg_2=obj.arg2())

    L("lazy-key-1")
    L("lazy-key-2", _path="content/file.ftl")
    L("lazy-key-3", arg_1="arg-1", arg_2="arg-2", _path="content/file.ftl")
    L("lazy-key-4", arg_1=arg1, arg_2=arg2)
    L("lazy-key-5", arg_1=obj.arg1, arg_2=obj.arg2)
    L("lazy-key-6", arg_1=obj.arg1(), arg_2=obj.arg2())

    i18n.attr.key.one()
    i18n.attr.key.two(_path="content/file.ftl")
    i18n.attr.key.three(arg_1="arg-1", arg_2="arg-2", _path="content/file.ftl")
    i18n.attr.key.four(arg_1=arg1, arg_2=arg2)
    i18n.attr.key.five(arg_1=obj.arg1, arg_2=obj.arg2)
    i18n.attr.key.six(arg_1=obj.arg1(), arg_2=obj.arg2())
"""


def test_common_extract(tmp_path: Path) -> None:
    (tmp_path / "test.py").write_text(CONTENT)

    fluent_keys_len = 24  # Number of keys in `CONTENT`.

    fluent_keys = extract_fluent_keys(
        path=tmp_path,
        i18n_keys=("i18n", "L", "LF"),
        ignore_attributes=IGNORE_ATTRIBUTES,
        ignore_kwargs=IGNORE_KWARGS,
        default_ftl_file=DEFAULT_FTL_FILE,
    )
    assert fluent_keys  # Check if `fluent_keys` is not empty.
    assert len(fluent_keys) == fluent_keys_len  # Check if `fluent_keys` has `fluent_keys_len` keys.
    assert "key-1" in fluent_keys
    assert "key-2" in fluent_keys
    assert "key-3" in fluent_keys
    assert "key-4" in fluent_keys
    assert "key-5" in fluent_keys
    assert "key-6" in fluent_keys
    assert "sugar_key_one" in fluent_keys
    assert "sugar_key_two" in fluent_keys
    assert "sugar_key_three" in fluent_keys
    assert "sugar_key_four" in fluent_keys
    assert "sugar_key_five" in fluent_keys
    assert "sugar_key_six" in fluent_keys
    assert "lazy-key-1" in fluent_keys
    assert "lazy-key-2" in fluent_keys
    assert "lazy-key-3" in fluent_keys
    assert "lazy-key-4" in fluent_keys
    assert "lazy-key-5" in fluent_keys
    assert "lazy-key-6" in fluent_keys

    assert fluent_keys["key-1"].key == "key-1"
    assert fluent_keys["key-1"].path == Path("_default.ftl")
    assert fluent_keys["key-1"].translation.value.elements[0].value == "key-1"
    assert fluent_keys["key-1"].code_path == tmp_path / "test.py"

    assert fluent_keys["key-2"].key == "key-2"
    assert fluent_keys["key-2"].path == Path("content/file.ftl")
    assert fluent_keys["key-2"].translation.value.elements[0].value == "key-2"
    assert fluent_keys["key-2"].code_path == tmp_path / "test.py"

    assert fluent_keys["key-3"].key == "key-3"
    assert fluent_keys["key-3"].path == Path("content/file.ftl")
    assert fluent_keys["key-3"].translation.value.elements[0].value == "key-3"
    assert fluent_keys["key-3"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["key-3"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["key-3"].code_path == tmp_path / "test.py"

    assert fluent_keys["key-4"].key == "key-4"
    assert fluent_keys["key-4"].path == Path("_default.ftl")
    assert fluent_keys["key-4"].translation.value.elements[0].value == "key-4"
    assert fluent_keys["key-4"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["key-4"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["key-4"].code_path == tmp_path / "test.py"

    assert fluent_keys["key-5"].key == "key-5"
    assert fluent_keys["key-5"].path == Path("_default.ftl")
    assert fluent_keys["key-5"].translation.value.elements[0].value == "key-5"
    assert fluent_keys["key-5"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["key-5"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["key-5"].code_path == tmp_path / "test.py"

    assert fluent_keys["key-6"].key == "key-6"
    assert fluent_keys["key-6"].path == Path("_default.ftl")
    assert fluent_keys["key-6"].translation.value.elements[0].value == "key-6"
    assert fluent_keys["key-6"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["key-6"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["key-6"].code_path == tmp_path / "test.py"

    assert fluent_keys["sugar_key_one"].key == "sugar_key_one"
    assert fluent_keys["sugar_key_one"].path == Path("_default.ftl")
    assert fluent_keys["sugar_key_one"].translation.value.elements[0].value == "sugar_key_one"
    assert fluent_keys["sugar_key_one"].code_path == tmp_path / "test.py"

    assert fluent_keys["sugar_key_two"].key == "sugar_key_two"
    assert fluent_keys["sugar_key_two"].path == Path("content/file.ftl")
    assert fluent_keys["sugar_key_two"].translation.value.elements[0].value == "sugar_key_two"
    assert fluent_keys["sugar_key_two"].code_path == tmp_path / "test.py"

    assert fluent_keys["sugar_key_three"].key == "sugar_key_three"
    assert fluent_keys["sugar_key_three"].path == Path("content/file.ftl")
    assert fluent_keys["sugar_key_three"].translation.value.elements[0].value == "sugar_key_three"
    assert (
        fluent_keys["sugar_key_three"].translation.value.elements[1].expression.id.name == "arg_1"
    )
    assert (
        fluent_keys["sugar_key_three"].translation.value.elements[2].expression.id.name == "arg_2"
    )
    assert fluent_keys["sugar_key_three"].code_path == tmp_path / "test.py"

    assert fluent_keys["sugar_key_four"].key == "sugar_key_four"
    assert fluent_keys["sugar_key_four"].path == Path("_default.ftl")
    assert fluent_keys["sugar_key_four"].translation.value.elements[0].value == "sugar_key_four"
    assert fluent_keys["sugar_key_four"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["sugar_key_four"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["sugar_key_four"].code_path == tmp_path / "test.py"

    assert fluent_keys["sugar_key_five"].key == "sugar_key_five"
    assert fluent_keys["sugar_key_five"].path == Path("_default.ftl")
    assert fluent_keys["sugar_key_five"].translation.value.elements[0].value == "sugar_key_five"
    assert fluent_keys["sugar_key_five"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["sugar_key_five"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["sugar_key_five"].code_path == tmp_path / "test.py"

    assert fluent_keys["sugar_key_six"].key == "sugar_key_six"
    assert fluent_keys["sugar_key_six"].path == Path("_default.ftl")
    assert fluent_keys["sugar_key_six"].translation.value.elements[0].value == "sugar_key_six"
    assert fluent_keys["sugar_key_six"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["sugar_key_six"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["sugar_key_six"].code_path == tmp_path / "test.py"

    assert fluent_keys["lazy-key-1"].key == "lazy-key-1"
    assert fluent_keys["lazy-key-1"].path == Path("_default.ftl")
    assert fluent_keys["lazy-key-1"].translation.value.elements[0].value == "lazy-key-1"
    assert fluent_keys["lazy-key-1"].code_path == tmp_path / "test.py"

    assert fluent_keys["lazy-key-2"].key == "lazy-key-2"
    assert fluent_keys["lazy-key-2"].path == Path("content/file.ftl")
    assert fluent_keys["lazy-key-2"].translation.value.elements[0].value == "lazy-key-2"
    assert fluent_keys["lazy-key-2"].code_path == tmp_path / "test.py"

    assert fluent_keys["lazy-key-3"].key == "lazy-key-3"
    assert fluent_keys["lazy-key-3"].path == Path("content/file.ftl")
    assert fluent_keys["lazy-key-3"].translation.value.elements[0].value == "lazy-key-3"
    assert fluent_keys["lazy-key-3"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["lazy-key-3"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["lazy-key-3"].code_path == tmp_path / "test.py"

    assert fluent_keys["lazy-key-4"].key == "lazy-key-4"
    assert fluent_keys["lazy-key-4"].path == Path("_default.ftl")
    assert fluent_keys["lazy-key-4"].translation.value.elements[0].value == "lazy-key-4"
    assert fluent_keys["lazy-key-4"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["lazy-key-4"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["lazy-key-4"].code_path == tmp_path / "test.py"

    assert fluent_keys["lazy-key-5"].key == "lazy-key-5"
    assert fluent_keys["lazy-key-5"].path == Path("_default.ftl")
    assert fluent_keys["lazy-key-5"].translation.value.elements[0].value == "lazy-key-5"
    assert fluent_keys["lazy-key-5"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["lazy-key-5"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["lazy-key-5"].code_path == tmp_path / "test.py"

    assert fluent_keys["lazy-key-6"].key == "lazy-key-6"
    assert fluent_keys["lazy-key-6"].path == Path("_default.ftl")
    assert fluent_keys["lazy-key-6"].translation.value.elements[0].value == "lazy-key-6"
    assert fluent_keys["lazy-key-6"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["lazy-key-6"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["lazy-key-6"].code_path == tmp_path / "test.py"

    assert fluent_keys["attr-key-one"].key == "attr-key-one"
    assert fluent_keys["attr-key-one"].path == Path("_default.ftl")
    assert fluent_keys["attr-key-one"].translation.value.elements[0].value == "attr-key-one"
    assert fluent_keys["attr-key-one"].code_path == tmp_path / "test.py"

    assert fluent_keys["attr-key-two"].key == "attr-key-two"
    assert fluent_keys["attr-key-two"].path == Path("content/file.ftl")
    assert fluent_keys["attr-key-two"].translation.value.elements[0].value == "attr-key-two"
    assert fluent_keys["attr-key-two"].code_path == tmp_path / "test.py"

    assert fluent_keys["attr-key-three"].key == "attr-key-three"
    assert fluent_keys["attr-key-three"].path == Path("content/file.ftl")
    assert fluent_keys["attr-key-three"].translation.value.elements[0].value == "attr-key-three"
    assert fluent_keys["attr-key-three"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["attr-key-three"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["attr-key-three"].code_path == tmp_path / "test.py"

    assert fluent_keys["attr-key-four"].key == "attr-key-four"
    assert fluent_keys["attr-key-four"].path == Path("_default.ftl")
    assert fluent_keys["attr-key-four"].translation.value.elements[0].value == "attr-key-four"
    assert fluent_keys["attr-key-four"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["attr-key-four"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["attr-key-four"].code_path == tmp_path / "test.py"

    assert fluent_keys["attr-key-five"].key == "attr-key-five"
    assert fluent_keys["attr-key-five"].path == Path("_default.ftl")
    assert fluent_keys["attr-key-five"].translation.value.elements[0].value == "attr-key-five"
    assert fluent_keys["attr-key-five"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["attr-key-five"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["attr-key-five"].code_path == tmp_path / "test.py"

    assert fluent_keys["attr-key-six"].key == "attr-key-six"
    assert fluent_keys["attr-key-six"].path == Path("_default.ftl")
    assert fluent_keys["attr-key-six"].translation.value.elements[0].value == "attr-key-six"
    assert fluent_keys["attr-key-six"].translation.value.elements[1].expression.id.name == "arg_1"
    assert fluent_keys["attr-key-six"].translation.value.elements[2].expression.id.name == "arg_2"
    assert fluent_keys["attr-key-six"].code_path == tmp_path / "test.py"


def test_extract_fluent_keys_no_files(tmp_path: Path) -> None:
    fluent_keys = extract_fluent_keys(
        path=tmp_path,
        i18n_keys="i18n",
        ignore_attributes=IGNORE_ATTRIBUTES,
        ignore_kwargs=IGNORE_KWARGS,
        default_ftl_file=DEFAULT_FTL_FILE,
    )
    assert not fluent_keys



from pathlib import Path

from fluent.syntax import ast

from ftl_extract.const import DEFAULT_FTL_FILE
from ftl_extract.matcher import FluentKey
from ftl_extract.process.kwargs_extractor import extract_kwargs


def extracts_variable_names_from_simple_variable() -> None:
    kwargs = extract_kwargs(
        FluentKey(
            code_path=Path("test.py"),
            key="key-1",
            translation=ast.Message(
                id=ast.Identifier("test_message"),
                value=ast.Pattern(
                    elements=[
                        ast.TextElement("Hello, "),
                        ast.Placeable(
                            expression=ast.VariableReference(id=ast.Identifier("username")),
                        ),
                    ],
                ),
            ),
            path=DEFAULT_FTL_FILE,
        ),
    )
    assert kwargs == {"username"}


def extracts_variable_names_from_select_expression() -> None:
    kwargs = extract_kwargs(
        FluentKey(
            code_path=Path("test.py"),
            key="key-1",
            translation=ast.Message(
                id=ast.Identifier("select_message"),
                value=ast.Pattern(
                    elements=[
                        ast.Placeable(
                            expression=ast.SelectExpression(
                                selector=ast.VariableReference(id=ast.Identifier("gender")),
                                variants=[
                                    ast.Variant(
                                        key=ast.Identifier("male"),
                                        value=ast.Pattern(elements=[ast.TextElement("Mr.")]),
                                        default=False,
                                    ),
                                    ast.Variant(
                                        key=ast.Identifier("female"),
                                        value=ast.Pattern(elements=[ast.TextElement("Ms.")]),
                                        default=False,
                                    ),
                                ],
                            ),
                        ),
                    ],
                ),
            ),
            path=DEFAULT_FTL_FILE,
        ),
    )
    assert kwargs == {"gender"}


def returns_empty_set_for_messages_without_variables() -> None:
    kwargs = extract_kwargs(
        FluentKey(
            code_path=Path("test.py"),
            key="key-1",
            translation=ast.Message(
                id=ast.Identifier("no_variables"),
                value=ast.Pattern(elements=[ast.TextElement("Just a text message.")]),
            ),
            path=DEFAULT_FTL_FILE,
        ),
    )
    assert kwargs == set()


def test_returns_empty_set_for_comment_translation() -> None:
    kwargs = extract_kwargs(
        FluentKey(
            code_path=Path("test.py"),
            key="key-1",
            translation=ast.Comment(content="This is a comment"),
            path=DEFAULT_FTL_FILE,
        ),
    )
    assert kwargs == set()


def test_returns_empty_set_for_translation_without_value() -> None:
    kwargs = extract_kwargs(
        FluentKey(
            code_path=Path("test.py"),
            key="key-1",
            translation=ast.Message(
                id=ast.Identifier("message_no_value"),
                value=None,  # Explicitly setting value to None
            ),
            path=DEFAULT_FTL_FILE,
        ),
    )
    assert kwargs == set()


def test_extracts_variable_names_from_mixed_elements() -> None:
    kwargs = extract_kwargs(
        FluentKey(
            code_path=Path("test.py"),
            key="key-1",
            translation=ast.Message(
                id=ast.Identifier("mixed_message"),
                value=ast.Pattern(
                    elements=[
                        ast.TextElement("Hello, "),
                        ast.Placeable(
                            expression=ast.VariableReference(id=ast.Identifier("username")),
                        ),
                        ast.TextElement(" your balance is "),
                        ast.Placeable(
                            expression=ast.VariableReference(id=ast.Identifier("balance")),
                        ),
                    ],
                ),
            ),
            path=DEFAULT_FTL_FILE,
        ),
    )
    assert kwargs == {"username", "balance"}


def test_extracts_selector_variable_name_from_select_expression() -> None:
    kwargs = extract_kwargs(
        FluentKey(
            code_path=Path("test.py"),
            key="key_select_expression",
            translation=ast.Message(
                id=ast.Identifier("select_expression_message"),
                value=ast.Pattern(
                    elements=[
                        ast.Placeable(
                            expression=ast.SelectExpression(
                                selector=ast.VariableReference(id=ast.Identifier("user_status")),
                                variants=[
                                    ast.Variant(
                                        key=ast.Identifier("active"),
                                        value=ast.Pattern(
                                            elements=[ast.TextElement("Active User")],
                                        ),
                                        default=False,
                                    ),
                                    ast.Variant(
                                        key=ast.Identifier("inactive"),
                                        value=ast.Pattern(
                                            elements=[ast.TextElement("Inactive User")],
                                        ),
                                        default=False,
                                    ),
                                ],
                            ),
                        ),
                    ],
                ),
            ),
            path=DEFAULT_FTL_FILE,
        ),
    )
    assert kwargs == {"user_status"}


def test_nested_extraction() -> None:
    kwargs = extract_kwargs(
        FluentKey(
            code_path=Path("test.py"),
            key="trade-waiting_for_answer",
            translation=ast.Message(
                id=ast.Identifier(name="nested-key"),
                value=ast.Pattern(
                    elements=[
                        ast.TextElement(value="nested-key\n"),
                        ast.Placeable(
                            expression=ast.SelectExpression(
                                selector=ast.VariableReference(
                                    id=ast.Identifier(name="first_level_key"),
                                ),
                                variants=[
                                    ast.Variant(
                                        key=ast.NumberLiteral(value="1"),
                                        value=ast.Pattern(
                                            elements=[
                                                ast.TextElement(value="✅ "),
                                                ast.Placeable(
                                                    expression=ast.SelectExpression(
                                                        selector=ast.VariableReference(
                                                            id=ast.Identifier(
                                                                name="second_level_key",
                                                            ),
                                                        ),
                                                        variants=[
                                                            ast.Variant(
                                                                key=ast.NumberLiteral(value="1"),
                                                                value=ast.Pattern(
                                                                    elements=[
                                                                        ast.TextElement(value="OK"),
                                                                    ],
                                                                ),
                                                            ),
                                                            ast.Variant(
                                                                key=ast.NumberLiteral(value="2"),
                                                                value=ast.Pattern(
                                                                    elements=[
                                                                        ast.TextElement(value="NO"),
                                                                    ],
                                                                ),
                                                            ),
                                                            ast.Variant(
                                                                key=ast.Identifier(name="other"),
                                                                value=ast.Pattern(
                                                                    elements=[
                                                                        ast.TextElement(
                                                                            value="ANOTHER",
                                                                        ),
                                                                    ],
                                                                ),
                                                                default=True,
                                                            ),
                                                        ],
                                                    ),
                                                ),
                                            ],
                                        ),
                                    ),
                                    ast.Variant(
                                        key=ast.NumberLiteral(value="0"),
                                        value=ast.Pattern(
                                            elements=[
                                                ast.TextElement(value="❌ "),
                                                ast.Placeable(
                                                    expression=ast.SelectExpression(
                                                        selector=ast.VariableReference(
                                                            id=ast.Identifier(
                                                                name="second_level_key",
                                                            ),
                                                        ),
                                                        variants=[
                                                            ast.Variant(
                                                                key=ast.NumberLiteral(value="1"),
                                                                value=ast.Pattern(
                                                                    elements=[
                                                                        ast.TextElement(value="OK"),
                                                                    ],
                                                                ),
                                                            ),
                                                            ast.Variant(
                                                                key=ast.NumberLiteral(value="2"),
                                                                value=ast.Pattern(
                                                                    elements=[
                                                                        ast.TextElement(value="NO"),
                                                                    ],
                                                                ),
                                                            ),
                                                            ast.Variant(
                                                                key=ast.Identifier(name="other"),
                                                                value=ast.Pattern(
                                                                    elements=[
                                                                        ast.TextElement(
                                                                            value="ANOTHER",
                                                                        ),
                                                                    ],
                                                                ),
                                                                default=True,
                                                            ),
                                                        ],
                                                    ),
                                                ),
                                            ],
                                        ),
                                    ),
                                    ast.Variant(
                                        key=ast.Identifier(name="other"),
                                        value=ast.Pattern(
                                            elements=[
                                                ast.TextElement(value="⏳ "),
                                                ast.Placeable(
                                                    expression=ast.VariableReference(
                                                        id=ast.Identifier(name="second_level_key"),
                                                    ),
                                                ),
                                                ast.TextElement(value=" ANOTHER"),
                                            ],
                                        ),
                                        default=True,
                                    ),
                                ],
                            ),
                        ),
                    ],
                ),
            ),
            path=DEFAULT_FTL_FILE,
            locale="en",
            position=0,
        ),
    )

    assert kwargs == {"first_level_key", "second_level_key"}



from pathlib import Path
from typing import Final

from fluent.syntax import FluentSerializer, parse

from ftl_extract.ftl_importer import import_from_ftl
from ftl_extract.process.commentator import comment_ftl_key
from ftl_extract.process.serializer import generate_ftl

CONTENT: Final[str] = """
key-1 = Key 1 {$var_reference_1} {$var_reference_2}
key-2 = Key 2 {msg_reference_1} {msg_reference_2}
key-3 = Key 3
key-4 = ⚙️ Header Text: { $selected_type ->
    [text] { set-type-text-button }
    [photo] { set-type-photo-button }
    [video] { set-type-video-button }
    [gif] { set-type-gif-button }
    [sticker] { set-type-sticker-button }
    *[unknown] 🤷‍♂️
}
key-5 = ⚠️ Header Text. Header Text.

    { chat_settings }

    💁‍♂️ Text inside <code>Continent/City</code>

    💡 Text inside:
        <blockquote><code>indent</code></blockquote>
      <blockquote><code>no indent</code></blockquote>
"""


def test_ftl_comment(tmp_path: Path) -> None:
    (tmp_path / "test.ftl").write_text(CONTENT, encoding="utf-8")

    ftl_keys, resource, leave_as_is = import_from_ftl(tmp_path / "test.ftl", "en")

    serializer = FluentSerializer(with_junk=True)

    comment_ftl_key(ftl_keys["key-1"], serializer=serializer)
    comment_ftl_key(ftl_keys["key-2"], serializer=serializer)
    comment_ftl_key(ftl_keys["key-3"], serializer=serializer)
    comment_ftl_key(ftl_keys["key-4"], serializer=serializer)
    comment_ftl_key(ftl_keys["key-5"], serializer=serializer)

    ftl, _ = generate_ftl(ftl_keys.values(), serializer=serializer, leave_as_is=leave_as_is)
    (tmp_path / "test.ftl").write_text(ftl, encoding="utf-8")

    ftl = (tmp_path / "test.ftl").read_text(encoding="utf-8")
    ftl = parse(ftl, with_spans=False)

    assert ftl.body[0].equals(ftl_keys["key-1"].translation)
    assert ftl.body[1].equals(ftl_keys["key-2"].translation)
    assert ftl.body[2].equals(ftl_keys["key-3"].translation)
    assert ftl.body[3].equals(ftl_keys["key-4"].translation)
    assert ftl.body[4].equals(ftl_keys["key-5"].translation)



from __future__ import annotations

from typing import TYPE_CHECKING

from click import echo
from fluent.syntax import FluentSerializer
from fluent.syntax import ast as fl_ast

from ftl_extract import extract_fluent_keys
from ftl_extract.code_extractor import sort_fluent_keys_by_path
from ftl_extract.const import DEFAULT_FTL_FILE, IGNORE_ATTRIBUTES, IGNORE_KWARGS
from ftl_extract.ftl_importer import import_ftl_from_dir
from ftl_extract.process.commentator import comment_ftl_key
from ftl_extract.process.kwargs_extractor import extract_kwargs
from ftl_extract.process.serializer import generate_ftl

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from ftl_extract.matcher import FluentKey


def extract(
    code_path: Path,
    output_path: Path,
    language: Iterable[str],
    i18n_keys: Iterable[str],
    ignore_attributes: str | Iterable[str] = IGNORE_ATTRIBUTES,
    expand_ignore_attributes: Iterable[str] | None = None,
    ignore_kwargs: str | Iterable[str] = IGNORE_KWARGS,
    comment_junks: bool = False,
    default_ftl_file: Path = DEFAULT_FTL_FILE,
    serializer: FluentSerializer | None = None,
) -> None:
    if expand_ignore_attributes is not None:
        ignore_attributes = frozenset(set(ignore_attributes) | set(expand_ignore_attributes or []))

    if serializer is None:
        serializer = FluentSerializer(with_junk=True)

    # Extract fluent keys from code
    in_code_fluent_keys = extract_fluent_keys(
        path=code_path,
        i18n_keys=i18n_keys,
        ignore_attributes=ignore_attributes,
        ignore_kwargs=ignore_kwargs,
        default_ftl_file=default_ftl_file,
    )

    for lang in language:
        # Import fluent keys from existing FTL files
        stored_fluent_keys, leave_as_is = import_ftl_from_dir(output_path, lang)
        for fluent_key in stored_fluent_keys.values():
            fluent_key.path = fluent_key.path.relative_to(output_path / lang)

        keys_to_comment: dict[str, FluentKey] = {}
        keys_to_add: dict[str, FluentKey] = {}

        # Find keys should be commented
        # Keys, that are not in code or not in their `path_` file
        # First step: find keys that have different paths
        for key, fluent_key in in_code_fluent_keys.items():
            if key in stored_fluent_keys and fluent_key.path != stored_fluent_keys[key].path:
                keys_to_comment[key] = stored_fluent_keys.pop(key)
                keys_to_add[key] = fluent_key

            elif key not in stored_fluent_keys:
                keys_to_add[key] = fluent_key

            else:
                stored_fluent_keys[key].code_path = fluent_key.code_path

        # Second step: find keys that have different kwargs
        for key, fluent_key in in_code_fluent_keys.items():
            if key not in stored_fluent_keys:
                continue

            fluent_key_placeable_set = extract_kwargs(fluent_key)
            stored_fluent_key_placeable_set = extract_kwargs(stored_fluent_keys[key])

            if fluent_key_placeable_set != stored_fluent_key_placeable_set:
                keys_to_comment[key] = stored_fluent_keys.pop(key)
                keys_to_add[key] = fluent_key

        # Third step: find keys that are not in code
        for key in stored_fluent_keys.keys() - in_code_fluent_keys.keys():
            keys_to_comment[key] = stored_fluent_keys.pop(key)

        for fluent_key in keys_to_comment.values():
            comment_ftl_key(fluent_key, serializer)

        # Comment Junk elements if needed
        if comment_junks is True:
            for fluent_key in leave_as_is:
                if isinstance(fluent_key.translation, fl_ast.Junk):
                    comment_ftl_key(fluent_key, serializer)

        sorted_fluent_keys = sort_fluent_keys_by_path(stored_fluent_keys)

        for path, keys in sort_fluent_keys_by_path(keys_to_add).items():
            sorted_fluent_keys.setdefault(path, []).extend(keys)

        for path, keys in sort_fluent_keys_by_path(keys_to_comment).items():
            sorted_fluent_keys.setdefault(path, []).extend(keys)

        leave_as_is_with_path: dict[Path, list[FluentKey]] = {}

        for fluent_key in leave_as_is:
            leave_as_is_with_path.setdefault(
                fluent_key.path.relative_to(output_path / lang),
                [],
            ).append(fluent_key)

        for path, keys in sorted_fluent_keys.items():
            ftl, _ = generate_ftl(
                keys,
                serializer=serializer,
                leave_as_is=leave_as_is_with_path.get(path, []),
            )
            (output_path / lang / path).parent.mkdir(parents=True, exist_ok=True)
            (output_path / lang / path).write_text(ftl, encoding="utf-8")
            echo(f"File {output_path / lang / path} has been saved. {len(keys)} keys updated.")



import ast
from pathlib import Path
from typing import cast
from unittest.mock import MagicMock, patch

import click.testing
import pytest
from click import BaseCommand
from fluent.syntax import FluentSerializer
from fluent.syntax import ast as fl_ast

from ftl_extract.cli import cli_extract
from ftl_extract.const import DEFAULT_FTL_FILE
from ftl_extract.ftl_extractor import extract
from ftl_extract.matcher import FluentKey, I18nMatcher


@pytest.fixture
def mock_fluent_key(tmp_path: Path) -> FluentKey:
    mock = MagicMock(spec=FluentKey)
    mock.code_path = tmp_path / "code"
    mock.translation = MagicMock(spec=fl_ast.Message)

    mock.translation.id = MagicMock(spec=fl_ast.Identifier)
    mock.translation.id.name = "key-1"

    text_element = MagicMock(spec=fl_ast.TextElement)
    text_element.value = "key-1"
    mock.translation.value = MagicMock(spec=fl_ast.Pattern)
    mock.translation.value.elements = [text_element]

    mock.translation.attributes = []

    mock.translation.comment = MagicMock(spec=fl_ast.Comment)
    mock.translation.comment.content = "Comment"

    return mock


@pytest.fixture
def setup_environment(tmp_path: Path) -> tuple[Path, Path]:
    code_path = tmp_path / "code"
    output_path = tmp_path / "output"
    code_path.mkdir()
    output_path.mkdir()
    return code_path, output_path


@pytest.fixture
def runner() -> click.testing.CliRunner:
    return click.testing.CliRunner()


@pytest.fixture
def mock_extract_function() -> patch:
    with patch("ftl_extract.cli.extract") as mock:
        yield mock


@pytest.fixture
def mock_leave_as_is() -> list:
    return [
        MagicMock(spec=fl_ast.Comment),
        MagicMock(spec=fl_ast.GroupComment),
        MagicMock(spec=fl_ast.ResourceComment),
        MagicMock(spec=fl_ast.Term),
        MagicMock(spec=fl_ast.Junk),
    ]


def test_extract_with_keys_to_comment_and_add(
    setup_environment: tuple[Path, Path],
    mock_fluent_key: FluentKey,
) -> None:
    code_path, output_path = setup_environment

    # Adjust the path to be relative to `output_path / "en"`, ensuring it's a valid subpath
    stored_fluent_key_path = (output_path / "en").resolve().joinpath("different/path.ftl")
    mock_fluent_key.path = Path("some/path.ftl")  # Path in code

    with (
        patch(
            "ftl_extract.ftl_extractor.extract_fluent_keys",
            return_value={"key-1": mock_fluent_key},
        ),
        patch(
            "ftl_extract.ftl_extractor.import_ftl_from_dir",
            return_value=({"key-1": MagicMock(spec=FluentKey, path=stored_fluent_key_path)}, []),
        ),
        patch("ftl_extract.ftl_extractor.comment_ftl_key") as mock_comment_ftl_key,
        patch(
            "ftl_extract.ftl_extractor.generate_ftl",
            return_value=("generated ftl", None),
        ) as mock_generate_ftl,
    ):
        extract(code_path, output_path, ("en",), ("i18n",))
        mock_comment_ftl_key.assert_called()
        mock_generate_ftl.assert_called()


def test_extract_with_keys_only_to_add(
    setup_environment: tuple[Path, Path],
    mock_fluent_key: FluentKey,
) -> None:
    code_path, output_path = setup_environment

    # Correctly set the path to be recognized as a subpath of `output_path / "en"`
    mock_fluent_key.path = output_path / "en" / "new" / "path.ftl"

    with (
        patch(
            "ftl_extract.ftl_extractor.extract_fluent_keys",
            return_value={"key-2": mock_fluent_key},
        ),
        patch(
            "ftl_extract.ftl_extractor.import_ftl_from_dir",
            return_value=({"key-1": mock_fluent_key}, []),
        ),
        patch(
            "ftl_extract.ftl_extractor.generate_ftl",
            return_value=("generated ftl", None),
        ) as mock_generate_ftl,
    ):
        extract(code_path, output_path, ("en",), ("i18n",))
        mock_generate_ftl.assert_called()


def test_extraction_with_valid_paths_succeeds(
    runner: click.testing.CliRunner,
    mock_extract_function: patch,
    tmp_path: Path,
) -> None:
    tmp_path.joinpath("path/to/code").mkdir(parents=True)
    code_path = tmp_path.joinpath("path/to/code")
    output_path = tmp_path.joinpath("path/to/output")

    result = runner.invoke(
        cast(BaseCommand, cli_extract),
        [code_path.as_posix(), output_path.as_posix()],
    )
    assert result.exit_code == 0
    assert f"Extracting from {code_path}..." in result.output
    mock_extract_function.assert_called_once()


def test_extraction_with_multiple_languages_handles_all(
    runner: click.testing.CliRunner,
    mock_extract_function: patch,
    tmp_path: Path,
) -> None:
    tmp_path.joinpath("path/to/code").mkdir(parents=True)
    code_path = tmp_path.joinpath("path/to/code")
    output_path = tmp_path.joinpath("path/to/output")

    result = runner.invoke(
        cast(BaseCommand, cli_extract),
        [code_path.as_posix(), output_path.as_posix(), "-l", "en", "-l", "fr"],
    )
    assert result.exit_code == 0
    assert mock_extract_function.call_args[1]["language"] == ("en", "fr")


def test_extraction_with_nonexistent_code_path_fails(runner: click.testing.CliRunner) -> None:
    result = runner.invoke(cast(BaseCommand, cli_extract), ["nonexistent/path", "path/to/output"])
    assert result.exit_code != 0
    assert "Invalid value for 'CODE_PATH'" in result.output


def test_extraction_with_invalid_i18n_keys_ignores_them(
    runner: click.testing.CliRunner,
    mock_extract_function: patch,
    tmp_path: Path,
) -> None:
    tmp_path.joinpath("path/to/code").mkdir(parents=True)
    code_path = tmp_path.joinpath("path/to/code")
    output_path = tmp_path.joinpath("path/to/output")

    result = runner.invoke(
        cast(BaseCommand, cli_extract),
        [code_path.as_posix(), output_path.as_posix(), "-k", "nonexistent_key"],
    )
    assert result.exit_code == 0
    assert mock_extract_function.call_args[1]["i18n_keys"] == ("nonexistent_key",)


def test_comment_junk_elements_if_needed(setup_environment: tuple[Path, Path]) -> None:
    code_path, output_path = setup_environment

    mock_junk_key = MagicMock(spec=FluentKey)
    mock_junk_key.translation = MagicMock(spec=fl_ast.Junk)
    mock_junk_key.path = MagicMock(spec=Path)
    mock_serializer = MagicMock(spec=FluentSerializer)

    with (
        patch("ftl_extract.ftl_extractor.extract_fluent_keys", return_value={}),
        patch("ftl_extract.ftl_extractor.import_ftl_from_dir", return_value=({}, [mock_junk_key])),
        patch("ftl_extract.ftl_extractor.comment_ftl_key") as mock_comment_ftl_key,
        patch("fluent.syntax.serializer.FluentSerializer", return_value=mock_serializer),
    ):
        extract(
            code_path,
            output_path,
            ("en",),
            ("i18n",),
            comment_junks=True,
            serializer=mock_serializer,
        )
        mock_comment_ftl_key.assert_called_once_with(mock_junk_key, mock_serializer)


def test_expand_ignore_attributes_updates_ignore_attributes(
    setup_environment: tuple[Path, Path],
) -> None:
    code_path, output_path = setup_environment
    initial_ignore_attributes = ["attr1", "attr2"]
    expand_ignore_attributes = ["attr3", "attr4"]
    expected_ignore_attributes = frozenset({"attr1", "attr2", "attr3", "attr4"})

    with (
        patch("ftl_extract.ftl_extractor.extract_fluent_keys", return_value={}),
        patch("ftl_extract.ftl_extractor.import_ftl_from_dir", return_value=({}, [])),
        patch("ftl_extract.ftl_extractor.comment_ftl_key"),
        patch("ftl_extract.ftl_extractor.generate_ftl", return_value=("generated ftl", None)),
    ):
        extract(
            code_path,
            output_path,
            ("en",),
            ("i18n",),
            ignore_attributes=initial_ignore_attributes,
            expand_ignore_attributes=expand_ignore_attributes,
        )

        assert (
            frozenset(initial_ignore_attributes) | frozenset(expand_ignore_attributes)
            == expected_ignore_attributes
        )


def test_stored_fluent_keys_code_path_update(setup_environment: tuple[Path, Path]) -> None:
    code_path, output_path = setup_environment
    mock_fluent_key = MagicMock(spec=FluentKey)
    mock_fluent_key.path = Path("_default.ftl")
    mock_fluent_key.code_path = code_path / "some_code_path.py"

    stored_fluent_key = MagicMock(spec=FluentKey)
    stored_fluent_key.path = Path(output_path / "en" / "_default.ftl")
    stored_fluent_key.code_path = None

    in_code_fluent_keys = {"key-1": mock_fluent_key}
    stored_fluent_keys = {"key-1": stored_fluent_key}

    with (
        patch("ftl_extract.ftl_extractor.extract_fluent_keys", return_value=in_code_fluent_keys),
        patch(
            "ftl_extract.ftl_extractor.import_ftl_from_dir",
            return_value=(stored_fluent_keys, []),
        ),
        patch("ftl_extract.ftl_extractor.extract_kwargs", return_value=set()),
        patch("ftl_extract.ftl_extractor.comment_ftl_key"),
        patch("ftl_extract.ftl_extractor.generate_ftl", return_value=("generated ftl", None)),
    ):
        extract(
            code_path,
            output_path,
            ("en",),
            ("i18n",),
        )

        assert stored_fluent_keys["key-1"].code_path == mock_fluent_key.code_path


def test_keys_to_comment_and_add_on_different_kwargs(setup_environment: tuple[Path, Path]) -> None:
    code_path, output_path = setup_environment
    mock_fluent_key = MagicMock(spec=FluentKey)
    mock_fluent_key.path = Path("_default.ftl")
    mock_fluent_key.code_path = code_path / "some_code_path.py"

    stored_fluent_key = MagicMock(spec=FluentKey)
    stored_fluent_key.path = Path(output_path / "en" / "_default.ftl")
    stored_fluent_key.code_path = None

    in_code_fluent_keys = {"key-1": mock_fluent_key}
    stored_fluent_keys = {"key-1": stored_fluent_key}

    with (
        patch("ftl_extract.ftl_extractor.extract_fluent_keys", return_value=in_code_fluent_keys),
        patch(
            "ftl_extract.ftl_extractor.import_ftl_from_dir",
            return_value=(stored_fluent_keys, []),
        ),
        patch("ftl_extract.ftl_extractor.extract_kwargs", side_effect=[{"arg1"}, {"arg2"}]),
        patch("ftl_extract.ftl_extractor.comment_ftl_key"),
        patch("ftl_extract.ftl_extractor.generate_ftl", return_value=("generated ftl", None)),
    ):
        extract(
            code_path,
            output_path,
            ("en",),
            ("i18n",),
        )

        assert "key-1" not in stored_fluent_keys
        assert "key-1" in in_code_fluent_keys
        assert in_code_fluent_keys["key-1"] == mock_fluent_key


def test_i18n_matcher_skips_call_with_no_args(setup_environment: tuple[Path, Path]) -> None:
    code_path, output_path = setup_environment
    matcher = I18nMatcher(code_path, default_ftl_file=DEFAULT_FTL_FILE)

    node = ast.Call(func=ast.Attribute(value=ast.Name(id="i18n"), attr="get"), args=[], keywords=[])
    matcher.visit_Call(node)

    assert len(matcher.fluent_keys) == 0


def test_generic_visit_called_on_else_block(setup_environment: tuple[Path, Path]) -> None:
    code_path, output_path = setup_environment
    matcher = I18nMatcher(code_path, default_ftl_file=DEFAULT_FTL_FILE)

    node = ast.Call(
        func=ast.Attribute(value=ast.Name(id="i18n"), attr="get"),
        args=[ast.Name(id="i18n")],
        keywords=[],
    )

    with patch.object(matcher, "generic_visit", wraps=matcher.generic_visit) as mock_generic_visit:
        matcher.visit_Call(node)
        mock_generic_visit.assert_called()


def test_generic_visit_called_when_attr_in_ignore_attributes(
    setup_environment: tuple[Path, Path],
) -> None:
    code_path, output_path = setup_environment
    matcher = I18nMatcher(
        code_path,
        default_ftl_file=DEFAULT_FTL_FILE,
        ignore_attributes={"ignore_this"},
    )

    # Create a mock AST node for a function call with an attribute in ignore_attributes
    node = ast.Call(
        func=ast.Attribute(
            value=ast.Name(id="i18n", ctx=ast.Load()),
            attr="ignore_this",
            ctx=ast.Load(),
        ),
        args=[ast.Constant(value="key")],
        keywords=[],
    )

    with patch.object(matcher, "generic_visit", wraps=matcher.generic_visit) as mock_generic_visit:
        matcher.visit_Call(node)
        mock_generic_visit.assert_called_with(node.args[0])

        assert len(matcher.fluent_keys) == 0


def test_i18n_matcher_skips_call_with_no_args_in_elif(setup_environment: tuple[Path, Path]) -> None:
    code_path, output_path = setup_environment
    matcher = I18nMatcher(code_path, default_ftl_file=DEFAULT_FTL_FILE)

    node = ast.Call(func=ast.Name(id="i18n", ctx=ast.Load()), args=[], keywords=[])
    matcher.visit_Call(node)

    assert len(matcher.fluent_keys) == 0


def test_i18n_matcher_ignore_kwargs(setup_environment: tuple[Path, Path]) -> None:
    code_path, output_path = setup_environment
    matcher = I18nMatcher(
        code_path,
        default_ftl_file=DEFAULT_FTL_FILE,
        ignore_kwargs={"ignore_this"},
    )

    node = ast.Call(
        func=ast.Attribute(
            value=ast.Name(id="i18n", ctx=ast.Load()),
            attr="get",
            ctx=ast.Load(),
        ),
        args=[ast.Constant(value="key")],
        keywords=[ast.keyword(arg="ignore_this", value=ast.Constant(value="value"))],
    )

    matcher.visit_Call(node)

    assert isinstance(matcher.fluent_keys["key"].translation.value.elements[0], fl_ast.TextElement)
    assert len(matcher.fluent_keys) == 1



from pathlib import Path
from typing import Final

import pytest

from ftl_extract.code_extractor import extract_fluent_keys
from ftl_extract.const import DEFAULT_FTL_FILE, IGNORE_ATTRIBUTES, IGNORE_KWARGS
from ftl_extract.exceptions import (
    FTLExtractorDifferentPathsError,
    FTLExtractorDifferentTranslationError,
)

CONTENT_1: Final[str] = """
# For `test_extract_similar_keys_in_different_paths_from_one_py_file` test.
def test(i18n):
    i18n.get("key-1", _path="content_2/file_1.ftl")
    i18n.get("key-1", _path="content_2/file_2.ftl")
"""

CONTENT_2_1: Final[str] = """
# For `test_extract_similar_fluent_keys_in_different_paths_from_different_py_files` test.
def test(i18n):
    i18n.get("key-1", arg_1="arg-1", _path="content_3/file_1.ftl")
"""

CONTENT_2_2: Final[str] = """
# For `test_extract_similar_fluent_keys_in_different_paths_from_different_py_files` test.
def test(i18n):
    i18n.get("key-1", arg_1="arg-1", _path="content_3/file_2.ftl")
"""
CONTENT_3: Final[str] = """
# For `test_extract_similar_fluent_keys_with_different_kwargs_one_py_file` test.
def test(i18n):
    i18n.get("key-1", arg_1="arg-1")
    i18n.get("key-1", arg_2="arg-2")
"""

CONTENT_4_1: Final[str] = """
# For `test_extract_similar_fluent_keys_with_different_kwargs_different_py_files` test.
def test(i18n):
    i18n.get("key-1", arg_1="arg-1")
"""

CONTENT_4_2: Final[str] = """
# For `test_extract_similar_fluent_keys_with_different_kwargs_different_py_files` test.
def test(i18n):
    i18n.get("key-1", arg_2="arg-2")
"""


def test_extract_similar_keys_in_different_paths_from_one_py_file(tmp_path: Path) -> None:
    (tmp_path / "test.py").write_text(CONTENT_1)

    with pytest.raises(FTLExtractorDifferentPathsError):
        extract_fluent_keys(
            path=tmp_path,
            i18n_keys="i18n",
            ignore_attributes=IGNORE_ATTRIBUTES,
            ignore_kwargs=IGNORE_KWARGS,
            default_ftl_file=DEFAULT_FTL_FILE,
        )


def test_extract_similar_fluent_keys_in_different_paths_from_different_py_files(
    tmp_path: Path,
) -> None:
    (tmp_path / "test.py").write_text(CONTENT_2_1)

    (tmp_path / "test2.py").write_text(CONTENT_2_2)

    with pytest.raises(FTLExtractorDifferentPathsError):
        extract_fluent_keys(
            path=tmp_path,
            i18n_keys="i18n",
            ignore_attributes=IGNORE_ATTRIBUTES,
            ignore_kwargs=IGNORE_KWARGS,
            default_ftl_file=DEFAULT_FTL_FILE,
        )


def test_extract_similar_fluent_keys_with_different_translation_one_py_file(tmp_path: Path) -> None:
    (tmp_path / "test.py").write_text(CONTENT_3)

    with pytest.raises(FTLExtractorDifferentTranslationError):
        extract_fluent_keys(
            path=tmp_path,
            i18n_keys="i18n",
            ignore_attributes=IGNORE_ATTRIBUTES,
            ignore_kwargs=IGNORE_KWARGS,
            default_ftl_file=DEFAULT_FTL_FILE,
        )


def test_extract_similar_fluent_keys_with_different_translation_different_py_files(
    tmp_path: Path,
) -> None:
    (tmp_path / "test.py").write_text(CONTENT_4_1)

    (tmp_path / "test2.py").write_text(CONTENT_4_2)

    with pytest.raises(FTLExtractorDifferentTranslationError):
        extract_fluent_keys(
            path=tmp_path,
            i18n_keys="i18n",
            ignore_attributes=IGNORE_ATTRIBUTES,
            ignore_kwargs=IGNORE_KWARGS,
            default_ftl_file=DEFAULT_FTL_FILE,
        )



from pathlib import Path
from unittest.mock import patch

import pytest
from fluent.syntax import ast

from ftl_extract.ftl_importer import import_from_ftl, import_ftl_from_dir


@pytest.fixture
def mock_ftl_content() -> str:
    return """
# Simple FTL file
## Group Comment
### Resource Comment
-term = Term value
hello = Hello, world!
welcome = Welcome, { $name }!
Junk
"""


def test_import_from_ftl_with_valid_ftl_file(mock_ftl_content: str) -> None:
    with patch("pathlib.Path.read_text", return_value=mock_ftl_content):
        keys, resource, leave_as_is = import_from_ftl(Path("/path/to/locale/en/example.ftl"), "en")
        assert "hello" in keys
        assert "welcome" in keys
        assert len(resource.body) == 7  # noqa: PLR2004


def test_import_from_ftl_with_empty_ftl_file() -> None:
    with patch("pathlib.Path.read_text", return_value=""):
        keys, resource, leave_as_is = import_from_ftl(Path("/path/to/locale/en/empty.ftl"), "en")
        assert len(keys) == 0
        assert len(resource.body) == 0


def test_import_ftl_from_dir_with_multiple_files(tmp_path: Path, mock_ftl_content: str) -> None:
    (tmp_path / "en").mkdir(parents=True)
    file1 = tmp_path / "en" / "file1.ftl"
    file2 = tmp_path / "en" / "file2.ftl"
    file1.write_text(mock_ftl_content)
    file2.write_text(mock_ftl_content)

    keys, leave_as_is = import_ftl_from_dir(tmp_path, "en")
    assert len(keys) == 2  # noqa: PLR2004


def test_import_ftl_from_dir_with_no_ftl_files(tmp_path: Path) -> None:
    (tmp_path / "en").mkdir(parents=True)
    keys, leave_as_is = import_ftl_from_dir(tmp_path, "en")
    assert len(keys) == 0


def test_import_ftl_from_dir_with_nonexistent_directory() -> None:
    with pytest.raises(FileNotFoundError):
        import_ftl_from_dir(Path("/path/to/nonexistent/dir"), "en")


def test_import_from_ftl_appends_non_message_entries_correctly(mock_ftl_content: str) -> None:
    with patch("pathlib.Path.read_text", return_value=mock_ftl_content):
        _, _, leave_as_is = import_from_ftl(Path("/path/to/locale/en/various_entries.ftl"), "en")
        assert len(leave_as_is) == 5  # noqa: PLR2004
        assert isinstance(leave_as_is[0].translation, ast.Comment)
        assert isinstance(leave_as_is[1].translation, ast.GroupComment)
        assert isinstance(leave_as_is[2].translation, ast.ResourceComment)
        assert isinstance(leave_as_is[3].translation, ast.Term)
        assert isinstance(leave_as_is[4].translation, ast.Junk)



from pathlib import Path

import pytest
from fluent.syntax import FluentSerializer, ast

from ftl_extract.const import DEFAULT_FTL_FILE
from ftl_extract.matcher import FluentKey
from ftl_extract.process.serializer import generate_ftl


@pytest.fixture
def single_fluent_key() -> list[FluentKey]:
    return [
        FluentKey(
            code_path=Path("test.py"),
            key="greeting",
            translation=ast.Message(
                id=ast.Identifier("greeting"),
                value=ast.Pattern(elements=[ast.TextElement("Hello, world!")]),
            ),
            path=DEFAULT_FTL_FILE,
        ),
    ]


@pytest.fixture
def multiple_fluent_keys() -> list[FluentKey]:
    return [
        FluentKey(
            code_path=Path("test.py"),
            key="greeting",
            translation=ast.Message(
                id=ast.Identifier("greeting"),
                value=ast.Pattern(elements=[ast.TextElement("Hello, world!")]),
            ),
            path=DEFAULT_FTL_FILE,
        ),
        FluentKey(
            code_path=Path("test.py"),
            key="farewell",
            translation=ast.Message(
                id=ast.Identifier("farewell"),
                value=ast.Pattern(elements=[ast.TextElement("Goodbye, world!")]),
            ),
            path=DEFAULT_FTL_FILE,
        ),
    ]


@pytest.fixture
def empty_fluent_keys() -> list[FluentKey]:
    return []


def test_custom_serializer_produces_correct_ftl_for_single_key(
    single_fluent_key: list[FluentKey],
) -> None:
    ftl_string, resource = generate_ftl(
        single_fluent_key,
        serializer=FluentSerializer(),
        leave_as_is=[],
    )
    assert "greeting = Hello, world!" in ftl_string
    assert len(resource.body) == 1


def test_custom_serializer_produces_correct_ftl_for_multiple_keys(
    multiple_fluent_keys: list[FluentKey],
) -> None:
    ftl_string, resource = generate_ftl(
        multiple_fluent_keys,
        serializer=FluentSerializer(),
        leave_as_is=[],
    )
    assert "greeting = Hello, world!" in ftl_string
    assert "farewell = Goodbye, world!" in ftl_string
    assert len(resource.body) == 2  # noqa: PLR2004


def test_custom_serializer_handles_empty_fluent_keys_list_properly(
    empty_fluent_keys: list[FluentKey],
) -> None:
    ftl_string, resource = generate_ftl(
        empty_fluent_keys,
        serializer=FluentSerializer(),
        leave_as_is=[],
    )
    assert ftl_string == ""
    assert resource.body is None or len(resource.body) == 0


def test_generate_ftl_includes_leave_as_is_elements() -> None:
    ftl_string, resource = generate_ftl(
        [
            FluentKey(
                code_path=Path("test.py"),
                key="test_key",
                translation=ast.Message(
                    id=ast.Identifier("test_message"),
                    value=ast.Pattern(elements=[ast.TextElement("Test message content")]),
                ),
                path=DEFAULT_FTL_FILE,
            ),
        ],
        serializer=FluentSerializer(with_junk=True),
        leave_as_is=[
            FluentKey(
                code_path=Path(),
                key="",
                translation=ast.Comment(content="This is a comment"),
                path=DEFAULT_FTL_FILE,
            ),
            FluentKey(
                code_path=Path(),
                key="",
                translation=ast.GroupComment(content="This is a group comment"),
                path=DEFAULT_FTL_FILE,
            ),
            FluentKey(
                code_path=Path(),
                key="",
                translation=ast.ResourceComment(content="This is a resource comment"),
                path=DEFAULT_FTL_FILE,
            ),
            FluentKey(
                code_path=Path(),
                key="",
                translation=ast.Term(
                    id=ast.Identifier("test_term"),
                    value=ast.Pattern(elements=[ast.TextElement("Test term content")]),
                ),
                path=DEFAULT_FTL_FILE,
            ),
            FluentKey(
                code_path=Path(),
                key="",
                translation=ast.Junk(
                    content="This is junk content",
                ),
                path=DEFAULT_FTL_FILE,
            ),
        ],
    )
    assert "This is a comment" in ftl_string
    assert "This is a group comment" in ftl_string
    assert "This is a resource comment" in ftl_string
    assert "test_term = Test term content" in ftl_string
    assert "This is junk content" in ftl_string
    assert "Test message content" in ftl_string



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Final, Literal


I18N_LITERAL: Final[Literal["i18n"]] = "i18n"
GET_LITERAL: Final[Literal["get"]] = "get"
PATH_LITERAL: Final[Literal["_path"]] = "_path"
IGNORE_ATTRIBUTES: Final[frozenset[str]] = frozenset(
    {"set_locale", "use_locale", "use_context", "set_context"},
)
IGNORE_KWARGS: Final[frozenset[str]] = frozenset()
DEFAULT_FTL_FILE: Final[Path] = Path("_default.ftl")



from __future__ import annotations

from typing import TYPE_CHECKING

from fluent.syntax import FluentSerializer, ast

if TYPE_CHECKING:
    from ftl_extract.matcher import FluentKey


def comment_ftl_key(key: FluentKey, serializer: FluentSerializer) -> None:
    raw_entry = serializer.serialize_entry(key.translation)
    key.translation = ast.Comment(content=raw_entry.strip())



from __future__ import annotations

from pathlib import Path
from time import perf_counter_ns

import click

from ftl_extract.const import DEFAULT_FTL_FILE, IGNORE_KWARGS
from ftl_extract.ftl_extractor import extract


@click.command()
@click.argument("code_path", type=click.Path(exists=True, path_type=Path))
@click.argument("output_path", type=click.Path(path_type=Path))
@click.option(
    "--language",
    "-l",
    multiple=True,
    default=("en",),
    show_default=True,
    help="Language of translation.",
)
@click.option(
    "--i18n_keys",
    "-k",
    default=("i18n", "L", "LazyProxy", "LazyFilter"),
    multiple=True,
    show_default=True,
    help="Names of function that is used to get translation.",
)
@click.option(
    "--ignore-attributes",
    default=("set_locale", "use_locale", "use_context", "set_context"),
    multiple=True,
    show_default=True,
    help="Ignore attributes, like `i18n.set_locale`.",
)
@click.option(
    "--expand-ignore-attributes",
    "-a",
    multiple=True,
    help="Expand default|targeted ignore attributes.",
)
@click.option(
    "--ignore-kwargs",
    default=IGNORE_KWARGS,
    multiple=True,
    show_default=True,
    help="Ignore kwargs, like `when` from `aiogram_dialog.I18nFormat(..., when=...)`.",
)
@click.option(
    "--comment-junks",
    is_flag=True,
    default=False,
    show_default=True,
    help="Comments Junk elements.",
)
@click.option(
    "--default-ftl-file",
    default=DEFAULT_FTL_FILE,
    show_default=True,
    type=click.Path(path_type=Path),
)
@click.version_option()
def cli_extract(
    code_path: Path,
    output_path: Path,
    language: tuple[str, ...],
    i18n_keys: tuple[str, ...],
    ignore_attributes: tuple[str, ...],
    expand_ignore_attributes: tuple[str, ...] | None = None,
    ignore_kwargs: tuple[str, ...] = (),
    comment_junks: bool = False,
    default_ftl_file: Path = DEFAULT_FTL_FILE,
) -> None:
    click.echo(f"Extracting from {code_path}...")
    start_time = perf_counter_ns()

    extract(
        code_path=code_path,
        output_path=output_path,
        language=language,
        i18n_keys=i18n_keys,
        ignore_attributes=ignore_attributes,
        expand_ignore_attributes=expand_ignore_attributes,
        ignore_kwargs=ignore_kwargs,
        comment_junks=comment_junks,
        default_ftl_file=default_ftl_file,
    )

    click.echo(f"Done in {(perf_counter_ns() - start_time) * 1e-9:.3f}s.")



from __future__ import annotations

import operator
from typing import TYPE_CHECKING

from fluent.syntax import FluentSerializer, ast

if TYPE_CHECKING:
    from collections.abc import Iterable

    from fluent.syntax.ast import Resource

    from ftl_extract.matcher import FluentKey


def generate_ftl(
    fluent_keys: Iterable[FluentKey],
    serializer: FluentSerializer,
    leave_as_is: Iterable[FluentKey],
) -> tuple[str, Resource]:
    """Generate FTL translations from `fluent_keys`."""
    resource = ast.Resource(body=None)

    listed_fluent_keys = list(fluent_keys)
    listed_fluent_keys.extend(leave_as_is)

    # Sort fluent keys by position
    listed_fluent_keys.sort(key=operator.attrgetter("position"))

    for fluent_key in listed_fluent_keys:
        resource.body.append(fluent_key.translation)

    return serializer.serialize(resource), resource



from pathlib import Path
from unittest.mock import Mock

from ftl_extract.code_extractor import post_process_fluent_keys
from ftl_extract.const import DEFAULT_FTL_FILE
from ftl_extract.matcher import FluentKey


def test_process_fluent_key() -> None:
    fluent_mock = Mock(spec=FluentKey)
    fluent_mock.path = "test.ftl"
    fluent_keys = {"key-1": fluent_mock}

    post_process_fluent_keys(fluent_keys, default_ftl_file=DEFAULT_FTL_FILE)
    assert fluent_mock.path == Path("test.ftl")


def test_process_fluent_key_default() -> None:
    fluent_mock = Mock(spec=FluentKey)
    fluent_mock.path = Path("test")
    fluent_keys = {"key-1": fluent_mock}

    post_process_fluent_keys(fluent_keys, default_ftl_file=DEFAULT_FTL_FILE)
    assert fluent_mock.path == Path("test/_default.ftl")
